
Program CURVeArrayFITprogram;

{By Will Polik, UC Berkeley, 11/8/86

 Fits (Xvector,Y) data points to an arbitrary function Functn}

{$N+}

uses DOS,CRT;

type
  Real = Double;

const
  MaxMatSize = 14;     {maximum number of fitting parameters}
  MaxVectorSize = 6;  {maximum dimension of independent variables}
  MaxDataSize = 250;    {maximum number of data points}
  Trm = '';

 type
   DataArray = Array[1..MaxDataSize] of Real;
   DataPointer = ^DataArray;
   Vector = Array[1..MaxVectorSize] of Real;
   VectorArray = Array[1..MaxDataSize] of Vector;
   VectorPointer = ^VectorArray;
   FitArray  = Array[1..MaxMatSize] of Real;
   InvMatrix = Array[1..MaxMatSize,1..MaxMatSize] of Real;

var
  X          : VectorPointer;
  Y,Yfit,W   : DataPointer;
  NPts,
  NParam     : Integer;
  Title,
  XLabel,
  Ylabel,
  OString    : String[80];
  A,SigmaA   : FitArray;
  ChiSqR     : Real;
  DataFile   : Text;
  i,j,
  Key        : Integer;
  WghtTyp    : Byte;     {0 = Equal, 1 = Relative, 2 = Absolute}
  NIter      : Integer;
  Code       : Integer;
  VectorDim  : Integer;
  ParDel,
  ChiDel     : Real;
  OutFile    : Text;

{$I C:\TP60\PASLIB\INFILE.PRO}
{$I C:\TP60\PASLIB\KEY.PRO}
{$I C:\TP60\PASLIB\SYMATINV.PRO}
{$I C:\TP60\PASLIB\OUTFILE.PRO}

(****************************************************************************)
(*                                                                          *)
(*    Code to be changed to allow a fit to a different function follows:    *)
(*                                                                          *)
(****************************************************************************)

  function Functn(X: Vector; A : FitArray) : Real;
  {Return Functn evaluated at X with parameters A}
  begin
  Functn := A[1] + A[2]*(X[1]+0.5) + A[3]*(X[2]+0.5) + A[4]*(X[1]+0.5)*(X[2]+0.5) +
            A[5]*SQR(X[1]+0.5) + A[6]*SQR(X[2]+0.5);
  end;

  procedure FDeriv(X: Vector; A : FitArray; Nparam: Integer;
                   var Deriv : FitArray);
  {Evaluate partial derivatives of Functn w.r.t. parameters A at X[i]}
  begin
  Deriv[1] := 1.0;
  Deriv[2] := X[1]+0.5;
  Deriv[3] := X[2]+0.5;
  Deriv[4] := (X[1]+0.5)*(X[2]+0.5);
  Deriv[5] := SQR(X[1]+0.5);
  Deriv[6] := SQR(X[2]+0.5);
  end;

  procedure FunctnOnce(A : FitArray; NParam: Integer);
  begin
  end;

  procedure FDerivOnce(A : FitArray; NParam: Integer);
  begin
  end;

  procedure InitialGuess;
  {Make initial estimates for all parameters; set fixed parameters
   (setting SigmaA[i] := -1 fixes A[i] and setting SigmaA[i] := 0
   frees A[i]); set convergence criteria}
  begin
  VectorDim := 2;
  Nparam := 6;
  A[1] := 9.5;
  A[2] := 0.0;
  A[3] := 0.0;
  A[4] := 0.0;
  A[5] := 0.0;
  A[6] := 0.0;
  SigmaA[1] := 0.0;
  SigmaA[2] := -1.0;
  SigmaA[3] := 0.0;
  SigmaA[4] := -1.0;
  SigmaA[5] := -1.0;
  SigmaA[6] := -1.0;
  ParDel := 1E-10;
  ChiDel := 1E-7;
  end;

(****************************************************************************)
(*                                                                          *)
(*       End of code to be changed for fitting to a different function.     *)
(*                                                                          *)
(*       You may want to experiment with the convergence criteria           *)
(*       factors, ParDel and ChiDel in InitialGuess, in order to achieve    *)
(*       acceptable convergence with the minimum number of iterations.      *)
(*                                                                          *)
(*       Finally, double-check your function and partial derivatives;       *)
(*       the program will bomb if they are incorrect!                       *)
(*                                                                          *)
(****************************************************************************)

{$I CURVAFIT.PRB}

procedure Initialize;
begin {of procedure Initialize}
ClrScr;
GotoXY(36,2); Write('CURVEARRAYFIT');
GotoXY(22,4); Write('by Will Polik, UC Berkeley, 11/8/86');
GotoXY(2,9); Write('A program which fits (Xvector,Y) data points to an arbitrary function');
GotoXY(22,11); Write('Data must exist in a file of the form:');
GotoXY(25,12); Write('Title');
GotoXY(25,13); Write('X Label');
GotoXY(25,14); Write('Y Label');
GotoXY(25,15); Write('Option String');
GotoXY(25,16); Write('X11    X12    X13  .  .   Y1    W1');
GotoXY(25,17); Write('X21    X22    X23  .  .   Y2    W2');
GotoXY(25,18); Write('X31    X32    X33  .  .   Y3    W3');
GotoXY(25,19); Write('.      .      .    .  .   .     . ');
GotoXY(25,20); Write('.      .      .    .  .   .     . ');
GotoXY(22,21); Write('where the weights (W) are optional');
GotoXY(14,25); Write('Strike Any Key to Proceed or <Control-C> to Exit...');
ClearKeyPressed; GetKeyPressed(Key);
New(X);
New(Y);
New(Yfit);
New(W);
InitialGuess;
end; {of procedure Initialize}

procedure GetData;
begin {of procedure GetData}
ClrScr;
OpenInFile(DataFile);
WriteLn;
Write('Are Weights Included? (y/n)');
Repeat
  ClearKeyPressed;
  GetKeyPressed(Key);
  Case Key of
    121,89 {y,Y} : WghtTyp := 1;
    110,78 {n,N} : WghtTyp := 0;
    Else Beep;
    end; {of case}
  Until Key in [121,89,110,78];
WriteLn;
if WghtTyp <> 0 then
begin
  WriteLn;
  Write('Relative or Absolute Weighting? (R/A)');
  Repeat
    ClearKeyPressed;
    GetKeyPressed(Key);
    Case Key of
      114,82 {r,R} : WghtTyp := 1;
      97,65  {a,A} : WghtTyp := 2;
    Else Beep;
    end; {of case}
  Until Key in [114,82,97,65];
end;
WriteLn;
WriteLn;
Npts := 0;
ReadLn(DataFile,Title);
ReadLn(DataFile,XLabel);
ReadLn(DataFile,YLabel);
ReadLn(DataFile,OString);
While not EOF(DataFile) do
  begin
  Npts := Npts + 1;
  if Npts > MaxDataSize
  then
    begin
    WriteLn('ERROR: Number of data points exceeds maximum limit of ',MaxDataSize);
    Close(DataFile);
    halt;
    end;
  for i := 1 to VectorDim do
    Read(DataFile,X^[NPts][i]);
  Read(DataFile,Y^[NPts]);
  if WghtTyp in [1,2]
  then
    begin
    Read(DataFile,W^[NPts]);
    if (WghtTyp=2) then W^[NPts] := 1/Sqr(W^[Npts]);
    end
  else W^[NPts] := 1;
  ReadLn(DataFile);
  end;
Close(DataFile);
end; {of procedure GetData}

procedure FitData;
begin {of procedure FitData}
InitialGuess;
CurveArrayFit(X,Y,W,NPts,NParam,A,SigmaA,ChiSqR,Yfit,Niter,Code,ParDel,ChiDel,True,WghtTyp);
WriteLn('The best fit of the data to the function after ',NIter,' iterations');
WriteLn('is given by:');
WriteLn;
for i := 1 to NParam do
  begin
  Write('A(',i,') = ',A[i]:16,'  +/-  ');
  if SigmaA[i] >= 0
  then WriteLn(SigmaA[i]:16)
  else WriteLn('fixed');
  end;
WriteLn;
WriteLn('The standard deviation [Sqrt(reduced chi squared)] = ',Sqrt(ChiSqR):16);
WriteLn;
Case Code of
  0:   WriteLn('Parameters converged');
  1:   WriteLn('Chi squared converged');
  10:  WriteLn('Unsuccessful Fit:  Flamda overflow');
  999: WriteLn('Unsuccessful Fit:  Other error');
  else WriteLn('Unsuccessful Fit:  Unknown return code');
  end; {of case}
WriteLn;
end; {of procedure FitData}

procedure OutputFit(var Device: Text);
begin {of procedure OutputFit}
WriteLn(Device,Title);
WriteLn(Device);
Write(Device,'  I       ');
for i := 1 to VectorDim do
  Write(Device,'X(I)(',i,')        ');
WriteLn(Device,'Y(I)          Yfit(I)     Yfit(I)-Y(I)       %DEV');
for i := 1 to NPts do
  begin
  Write(Device,i:3,'   ');
  for j := 1 to VectorDim do
    Write(Device,X^[i][j]:12,'   ');
  Write(Device,Y^[i]:12,'   ',Yfit^[i]:12,'   ',Yfit^[i]-Y^[i]:12);
  if Y^[i] <> 0
  then WriteLn(Device,'   ',100.0*(Yfit^[i]-Y^[i])/Abs(Y^[i]):10:4)
  else WriteLn(Device);
  end;
WriteLn(Device);
WriteLn(Device,'The best fit of the data is achieved with:');
WriteLn(Device);
for i := 1 to NParam do
  WriteLn(Device,'A(',i-1,') = ',A[i]:16,'  +/-  ',SigmaA[i]:16);
WriteLn(Device);
WriteLn(Device,'The standard deviation [Sqrt(reduced chi-square)] = ',Sqrt(ChiSqR):16);
end; {of procedure OutputFit}

(*  Cannot output an (X,Y) list because X is a vector !!!
procedure OutputList(var Device: Text);
var
  Start,Stop,Step,Temp : Real;
begin {of procedure OutputList}
Write('Input Start Value: '); ReadLn(Start);
Write('Input Stop Value:  '); ReadLn(Stop);
Write('Input Step Value:  '); ReadLn(Step);
WriteLn;
if Start > Stop
then
  begin
  Temp := Start;
  Start := Stop;
  Stop := Temp;
  end;
WriteLn(Device,Title);
WriteLn(Device,XLabel);
WriteLn(Device,Ylabel);
WriteLn(Device,'/L');
Repeat
  WriteLn(Device,Start:12,'   ',Functn(Start,A):12);
  Start := Start + Abs(Step);
  Until Start > Stop;
end; {of procedure OutputList}
*)

procedure DisplayFit;
begin {of procedure DisplayFit}
Write('Would you like to view a listing of the data and the fit? (y/n)');
ClearKeyPressed; GetKeyPressed(Key);
WriteLn; WriteLn;
if Key in [121,89] {y,Y}
then
  begin
  AssignCRT(OutFile);
  Rewrite(OutFile);
  OutputFit(OutFile);
  Close(OutFile);
  WriteLn;
  end;
(*
Write('Would you like to view a step-wise listing of the fitting function? (y/n)');
ClearKeyPressed; GetKeyPressed(Key);
WriteLn; WriteLn;
if Key in [121,89] {y,Y}
then
  begin
  AssignCRT(OutFile);
  Rewrite(OutFile);
  OutputFit(OutFile);
  Close(OutFile);
  WriteLn;
  end;
*)
end; {of procedure DisplayFit}

procedure FileFit;
begin {of procedure FileFit}
Write('Would you like to file a listing of the data and the fit? (y/n)');
ClearKeyPressed; GetKeyPressed(Key);
WriteLn; WriteLn;
if Key in [121,89] {y,Y}
then
  begin
  OpenOutFile(OutFile);
  WriteLn;
  OutputFit(OutFile);
  Close(OutFile);
  end;
(*
Write('Would you like to file a step-wise listing of the fitting function? (y/n)');
ClearKeyPressed; GetKeyPressed(Key);
WriteLn; WriteLn;
if Key in [121,89] {y,Y}
then
  begin
  OpenOutFile(OutFile);
  WriteLn;
  OutputList(OutFile);
  Close(OutFile);
  end;
*)
end; {of procedure FileFit}

begin {of program CurveFit}
Initialize;
GetData;
FitData;
DisplayFit;
FileFit;
end. {of program CurveFit}
