
procedure SymMatInv(var Matrix: InvMatrix; N: Integer; var Determinant: Double);

{Inverts a symmetric matrix and calculates its determinant

 by Will Polik, UC Berkeley, 5/6/86

 Adapted from Bevington, Data Reduction and Error Analysis for the Physical
 Sciences, McGraw-Hill, New York, 1969, p. 302

 Assumes:
   const
     MaxMatSize = ??;
   type
     InvMatrix = Array[1..MaxMatSize,1..MaxMatSize] of Real;

 Parameters:
   Matrix      - Input matrix which is replaced by its inverse
   N           - Degree of matrix
   Determinant - Determinant of Input Matrix}

label
  TryAgain;

var
  I,J,K,L   : Integer;
  Amax,Save : Real;
  IK,JK     : Array[1..MaxMatSize] of Integer;

begin {of procedure SymMatInv}
if N > MaxMatSize
then
  begin
  WriteLn('ERROR: N = ',N,' > MaxMatSize = ',MaxMatSize,' in SymMatInv');
  halt;
  end;
Determinant := 1.0;
for K := 1 to N do
  begin
  {find largest element in rest of matrix}
  Amax := 0;
  TryAgain:
  for I := K to N do
    begin
    for J := K to N do
      if Abs(Amax) <= Abs(Matrix[I,J])
      then
        begin
        Amax := Matrix[I,J];
        IK[K] := I;
        JK[K] := J;
        end;
    end;
  {Interchange rows and columns to put Amax in Matrix[K,K]}
  if Amax = 0
  then
    begin
    Determinant := 0;
    exit;
    end;
  I := IK[K];
  if K > I then goto TryAgain;
  if I > K
  then for J := 1 to N do
    begin
    Save := Matrix[K,J];
    Matrix[K,J] := Matrix[I,J];
    Matrix[I,J] := -Save;
    end;
  J := JK[K];
  if K > J then goto TryAgain;
  if J > K
  then for I := 1 to N do
    begin
    Save := Matrix[I,K];
    Matrix[I,K] := Matrix[I,J];
    Matrix[I,J] := -Save;
    end;
  {Accumulate elements of inverse matrix}
  for I := 1 to N do
    if I <> K
    then Matrix[I,K] := -Matrix[I,K]/Amax;
  for I := 1 to N do
    begin
    for J := 1 to N do
      if (I <> K) and (J <> K)
      then Matrix[I,J] := Matrix[I,J] + Matrix[I,K] * Matrix[K,J];
    end;
  for J := 1 to N do
    if J <> K
    then Matrix[K,J] := Matrix[K,J]/Amax;
  Matrix[K,K] := 1.0/Amax;
  Determinant := Determinant * Amax;
  end;
{Restore ordering of matrix}
for L := 1 to N do
  begin
  K := N - L + 1;
  J := IK[K];
  if J > K
  then for I := 1 to N do
    begin
    Save := Matrix[I,K];
    Matrix[I,K] := -Matrix[I,J];
    Matrix[I,J] := Save;
    end;
  I := JK[K];
  if I > K
  then for J := 1 to N do
    begin
    Save := Matrix[K,J];
    Matrix[K,J] := -Matrix[I,J];
    Matrix[I,J] := Save;
    end;
  end;
end; {of procedure SymMatInv}