
{Real symmetric matrix diagonalization and sorting procedures:

SYMEIG(A:glnpnp; D,E:glnp; N:Integer; Vect:Boolean);

     Input:  A    - N by N matrix to be diagonalized
             N    - logical dimension of A
             Vect - True if eigenvectors are desired; False if only
                    eigenvalues are desired

     Output: A    - matrix of eigenvectors (if Vect = True)
             D    - vector of eigenvalues

     Note:   E is temporary storage (neither input nor output)
             Eigenvalue D[i] corresponds to eigenvector A[j,i]

EIGSORT(A:glnpnp; D:glnp; N:Integer);

     Input:  A - N by N matrix of eigenvectors
             D - vector of eigenvalues
             N - logical dimension of A

     Output: A - N by N matrix of sorted eigenvectors
             D - vector of eigenvalues sorted into ascending order

     Note:   Eigenvalue D[i] corresponds to eigenvector A[j,i]

Programs using these routines must contain:

Const
  np = ???;
Type
   glnp = ARRAY [1..np] OF real;
   glnpnp = ARRAY [1..np,1..np] OF real;

where 'np by np' is the physical dimension of the matrix to be analyzed.}

procedure symeig(var a: glnpnp; var d,e: glnp; n: Integer; FindVectors: Boolean);

PROCEDURE tred2(VAR a: glnpnp; n: integer; VAR d,e: glnp);
VAR
   l,k,j,i: integer;
   scale,hh,h,g,f: real;
FUNCTION sign(a,b: real): real;
   BEGIN
      IF (b < 0) THEN sign := -abs(a) ELSE sign := abs(a)
   END;
BEGIN
   IF (n > 1) THEN BEGIN
      FOR i := n DOWNTO 2 DO BEGIN
         l := i-1;
         h := 0.0;
         scale := 0.0;
         IF (l > 1) THEN BEGIN
            FOR k := 1 to l DO BEGIN
               scale := scale+abs(a[i,k])
            END;
            IF (scale = 0.0) THEN BEGIN
               e[i] := a[i,l]
            END ELSE BEGIN
               FOR k := 1 to l DO BEGIN
                  a[i,k] := a[i,k]/scale;
                  h := h+sqr(a[i,k])
               END;
               f := a[i,l];
               g := -sign(sqrt(h),f);
               e[i] := scale*g;
               h := h-f*g;
               a[i,l] := f-g;
               f := 0.0;
               FOR j := 1 to l DO BEGIN
                  IF FindVectors THEN BEGIN
                     a[j,i] := a[i,j]/h;
                  END;
                  g := 0.0;
                  FOR k := 1 to j DO BEGIN
                     g := g+a[j,k]*a[i,k]
                  END;
                  IF (l > j) THEN FOR k := j+1 to l DO g := g+a[k,j]*a[i,k];
                  e[j] := g/h;
                  f := f+e[j]*a[i,j]
               END;
               hh := f/(h+h);
               FOR j := 1 to l DO BEGIN
                  f := a[i,j];
                  g := e[j]-hh*f;
                  e[j] := g;
                  FOR k := 1 to j DO a[j,k] := a[j,k]-f*e[k]-g*a[i,k]
               END
            END
         END ELSE BEGIN
            e[i] := a[i,l]
         END;
         d[i] := h
      END
   END;
   IF FindVectors THEN BEGIN
      d[1] := 0.0;
   END;
   e[1] := 0.0;
   FOR i := 1 to n DO BEGIN
      IF FindVectors THEN BEGIN
         l := i-1;
         IF (d[i] <> 0.0) THEN BEGIN
            FOR j := 1 to l DO BEGIN
               g := 0.0;
               FOR k := 1 to l DO BEGIN
                  g := g+a[i,k]*a[k,j]
               END;
               FOR k := 1 to l DO BEGIN
                  a[k,j] := a[k,j]-g*a[k,i]
               END
            END
         END;
         d[i] := a[i,i];
         a[i,i] := 1.0;
         IF (l >= 1) THEN BEGIN
            FOR j := 1 to l DO BEGIN
               a[i,j] := 0.0;
               a[j,i] := 0.0;
            END
          END
      END ELSE BEGIN
        d[i] := a[i,i];
      END
   END
END;

PROCEDURE tqli(VAR d,e: glnp; n: integer; VAR z: glnpnp);
LABEL 1,2;
VAR
   m,l,iter,i,k: integer;
   s,r,p,g,f,dd,c,b: real;
FUNCTION sign(a,b: real): real;
   BEGIN
      IF (b < 0) THEN sign := -abs(a) ELSE sign := abs(a)
   END;
BEGIN
   IF  (n > 1)  THEN BEGIN
      FOR i := 2 to n DO BEGIN
         e[i-1] := e[i]
      END;
      e[n] := 0.0;
      FOR l := 1 to n DO BEGIN
         iter := 0;
1:         FOR m := l to n-1 DO BEGIN
            dd := abs(d[m])+abs(d[m+1]);
            IF  (abs(e[m])+dd = dd) THEN  GOTO 2
         END;
         m := n;
2:         IF (m <> l) THEN BEGIN
            IF (iter = 30) THEN BEGIN
               writeln('pause in routine TQLI');
               writeln('too many iterations'); readln
            END;
            iter := iter+1;
            g := (d[l+1]-d[l])/(2.0*e[l]);
            r := sqrt(sqr(g)+1.0);
            g := d[m]-d[l]+e[l]/(g+sign(r,g));
            s := 1.0;
            c := 1.0;
            p := 0.0;
            FOR i := m-1 DOWNTO l DO BEGIN
               f := s*e[i];
               b := c*e[i];
               IF (abs(f) >= abs(g)) THEN BEGIN
                  c := g/f;
                  r := sqrt(sqr(c)+1.0);
                  e[i+1] := f*r;
                  s := 1.0/r;
                  c := c*s
               END ELSE BEGIN
                  s := f/g;
                  r := sqrt(sqr(s)+1.0);
                  e[i+1] := g*r;
                  c := 1.0/r;
                  s := s*c
               END;
               g := d[i+1]-p;
               r := (d[i]-g)*s+2.0*c*b;
               p := s*r;
               d[i+1] := g+p;
               g := c*r-b;
            (* Next loop can be omitted if eigenvectors not wanted *)
               IF FindVectors THEN BEGIN
                  FOR k := 1 to n DO BEGIN
                     f := z[k,i+1];
                     z[k,i+1] := s*z[k,i]+c*f;
                     z[k,i] := c*z[k,i]-s*f
                  END
               END;
            END;
            d[l] := d[l]-p;
            e[l] := g;
            e[m] := 0.0;
            GOTO 1
         END
      END
   END
END;

begin {of Procedure SYMEIG}
tred2(a,n,d,e);
tqli(d,e,n,a);
end; {of Procedure SYMEIG}

PROCEDURE eigsrt(VAR v: glnpnp; VAR d: glnp; n: integer);
VAR
   k,j,i: integer;
   p: real;
BEGIN
   FOR i := 1 to n-1 DO BEGIN
      k := i;
      p := d[i];
      FOR j := i+1 to n DO BEGIN
         IF (d[j] <= p) THEN BEGIN
            k := j;
            p := d[j]
         END
      END;
      IF (k <> i) THEN BEGIN
         d[k] := d[i];
         d[i] := p;
         FOR j := 1 to n DO BEGIN
            p := v[j,i];
            v[j,i] := v[j,k];
            v[j,k] := p
         END
      END
   END
END;
